"""Test the PQCA class."""

# pylint: disable=import-error
import qiskit
import pqca
import pqca.backend


def test_create_automaton():
    """1-D automaton, no simulation."""
    cx_circuit = qiskit.QuantumCircuit(2)
    cx_circuit.cx(0, 1)
    tes = pqca.tessellation.one_dimensional(10, 2)
    automaton = pqca.Automaton(
        [0]*10, [pqca.UpdateFrame(tes, qiskit_circuit=cx_circuit)], lambda x: [1]*10)
    assert automaton.iterate() == [[1]*10]
    assert len(str(automaton)) > 0


def test_aer_backend():
    """Simulate 1-D automaton."""
    cx_circuit = qiskit.QuantumCircuit(2)
    cx_circuit.cx(0, 1)
    tes = pqca.tessellation.one_dimensional(4, 2)

    def backend(circuit):
        return pqca.backend.Aer(circuit, "qasm_simulator")

    automaton = pqca.Automaton(
        [1]*4, [pqca.UpdateFrame(tes, qiskit_circuit=cx_circuit)], backend)
    assert automaton.iterate() == [[1, 0]*2]



"""An UpdateFrame holds cell-circuit and tessellation data."""

import re
from typing import List
from qiskit import QuantumCircuit
from qiskit.circuit.quantumregister import Qubit, QuantumRegister
from .tessellation import Tessellation
from .exceptions import (TooLittleDataForUpdateFrame,
                         TooMuchDataForUpdateFrame,
                         CircuitWrongShapeForCell)


class UpdateFrame:
    """Create a large circuit from a tessellated small circuit."""

    cell_circuit: QuantumCircuit
    tessellation: Tessellation
    full_circuit_instructions: List

    def __init__(self,
                 tessellation: Tessellation,
                 qasm_circuit_file: str = None,
                 qasm_data_as_string: str = None,
                 qiskit_circuit: QuantumCircuit = None):
        """Hold the circuit to be applied to each cell in the tessellation."""
        count_arguments_not_none: int = len([arg for arg in (
            qasm_circuit_file, qasm_data_as_string, qiskit_circuit) if arg is not None])
        if count_arguments_not_none == 0:
            raise TooLittleDataForUpdateFrame
        elif count_arguments_not_none > 1:
            raise TooMuchDataForUpdateFrame

        if qasm_data_as_string is not None:
            self.cell_circuit = QuantumCircuit.from_qasm_str(
                qasm_data_as_string)
        if qasm_circuit_file is not None:
            self.cell_circuit = QuantumCircuit.from_qasm_file(
                qasm_circuit_file)
        if qiskit_circuit is not None:
            self.cell_circuit = qiskit_circuit

        self.tessellation = tessellation
        self.full_circuit_instructions = wind_circuit_around_loop(
            self.cell_circuit, self.tessellation)

    def __str__(self):
        """Human-readable string representation."""
        return f"UpdateFrame(circuit: {self.cell_circuit} on each cell of {str(self.tessellation)})"


def wind_circuit_around_loop(circuit: QuantumCircuit, tessellation: Tessellation):
    """Return the tessellated circuit as a list of instructions."""
    if len(circuit.qubits) > len(tessellation.cells[0]):
        raise CircuitWrongShapeForCell(
            circuit.qubits, len(tessellation.cells[0]))
    next_column = []

    qreg = QuantumRegister(tessellation.size, circuit.qregs[0].name)

    for cell in tessellation.cells:
        for instruction, qargs, cargs in circuit.data:
            # Create a new instruction pointing at the correct qubits
            instruction_context = instruction, [
                Qubit(qreg, cell[q.index]) for q in qargs], cargs
            next_column.append(instruction_context)
    return next_column



"""Create a tessellation from a list of qubits."""

from __future__ import annotations
from typing import List, Callable, Tuple, Iterable
import itertools
import functools
from . import exceptions
from .vector import Vector


class Tessellation:
    """Describes how to partition the lattice into cells.

    This is a list of lists of qubits that also performs some validation.
    """

    size: int

    def __init__(self, cells: List[int]):
        """Treat a list of cells as a tessellation."""
        if len(cells) == 0:
            raise exceptions.NoCellsException
        if len(cells[0]) == 0:
            raise exceptions.EmptyCellException

        self.cells = cells
        self.size = sum([len(c) for c in cells])

        # check no qubit appears twice
        set_of_qubits = {q for cell in self.cells for q in cell}
        if len(set_of_qubits) != self.size:
            raise exceptions.PartitionUnevenlyCoversQubits(cells)
        # check all cells the same size
        if len({len(cell) for cell in cells}) != 1:
            raise exceptions.IrregularCellSize(cells)

    def shifted_by(self, amount=1):
        """Shift the tessellation along by the specified amount."""
        return self.update_names(lambda x: x+amount)

    def update_names(self,
                     name_update: Callable[[int], int],
                     rename_modulo_size=True) -> Tessellation:
        """Create a new tessellation by renaming every qubit in the old one.

        This renaming will happen modulo the number of qubits unless asked not to.
        """
        if rename_modulo_size:
            def make_address_positive(qubit_name):
                while qubit_name < 0:
                    qubit_name += self.size
                while qubit_name >= self.size:
                    qubit_name -= self.size
                return qubit_name
            return Tessellation([
                [make_address_positive(name_update(c)) for c in cell] for cell in self.cells])
        else:
            return Tessellation([[name_update(c) for c in cell] for cell in self.cells])

    def __str__(self):
        """Human-readable string representation."""
        return f"Tessellation({self.size} qubits as " +\
            f"{len(self.cells)} cells, first cell: {self.cells[0]})"


def one_dimensional(num_qubits: int, cell_size: int) -> Tessellation:
    """Partition a line of length num_qubits into cells of size cell_size."""
    return n_dimensional([num_qubits], [cell_size])


def cell_of_given_size(cell_dimension: List[int]) -> List[Vector]:
    """Given a length in each dimension create a cell."""
    points_in_first_cell = itertools.product(
        *list(map(lambda d: list(range(0, d)), cell_dimension))
    )

    first_cell_as_vectors = map(lambda t: Vector(
        list(t)), list(points_in_first_cell))
    return list(first_cell_as_vectors)


def n_dimensional(qubits_in_each_dimension: List[int],
                  cell_size: List[int]) -> Tessellation:
    """Partition an n-dimensional lattice into n-dimensional cuboids.

    example qubits_in_each_dimension: [5,5,10] for a cuboid of size 5 by 5 by 10
    example cell_size: [5,5,2]

    Note that the size of the cell must divide (for each dimension) the number
    of qubits in the lattice.
    """
    for index, length in enumerate(qubits_in_each_dimension):
        try:
            assert length % cell_size[index] == 0
        except:
            raise exceptions.IrregularCoordinateDimensions(
                qubits_in_each_dimension, cell_size) from None

    focal_points_as_product = itertools.product(
        *[range(0, qubits_in_each_dimension[index], cell_size[index])
          for index, _ in enumerate(qubits_in_each_dimension)]
    )

    # "Starting" point in each cell
    focal_points = map(Vector, list(focal_points_as_product))

    # Points in first cell
    points_in_first_cell = cell_of_given_size(cell_size)

    # List of list of vectors
    cells_as_vectors = [[focal_point + delta for delta in points_in_first_cell]
                        for focal_point in focal_points]

    # List of list of names (as integers)
    cells = [[vector_to_name(qubit_vector, qubits_in_each_dimension) for qubit_vector in cell]
             for cell in cells_as_vectors]

    return Tessellation(cells)


def vector_to_name(qubit_vector: Vector, qubits_in_each_dimension: List[int]):
    """Find the point in a lexicographic order for a vector.

    This turns a vector in the lattice into an unique "name" (int),
    allowing a linear collection of qubits to represent a lattice.
    """
    def sum_weighted_lengths(index, length, acc):
        """From an initial segment of qubits_in_each_dimension make a weighting."""
        def multiply(acc, entry):
            return entry*acc
        return acc + length * functools.reduce(multiply, qubits_in_each_dimension[index+1:], 1)

    def unpack(acc, index_entry):
        return sum_weighted_lengths(index_entry[0], index_entry[1], acc)

    return functools.reduce(unpack, enumerate(qubit_vector.entries), 0)


__all__ = ["Tessellation", "one_dimensional", "n_dimensional"]



"""Expose ways of evaluating circuits."""

import qiskit

from . import exceptions


def Aer(circuit: qiskit.QuantumCircuit, qiskit_backend="qasm_simulator"):
    """Expose the qiskit Aer Simulator."""
    circuit.measure_all()
    results = qiskit.execute(circuit, qiskit.Aer.get_backend(
        qiskit_backend
    ), shots=1).result()
    if results.success:
        final_state_as_string = list(results.get_counts(circuit).keys())[0]
        return [int(x) for x in final_state_as_string[::-1]]
    raise exceptions.BackendError(results.status)



"""Exceptions for the PQCA module."""


class PQCAException(Exception):
    """Base exception class for the pqca module."""


class TooLittleDataForUpdateFrame(PQCAException):
    """Too few arguments given."""

    def __init__(self):
        """Create exception."""
        super().__init__("No circuit, in any format, was provided.")


class TooMuchDataForUpdateFrame(PQCAException):
    """Too many arguments given."""

    def __init__(self):
        """Create exception."""
        super().__init__("Data given in multiple formats; need exactly one.")


class PartitionUnevenlyCoversQubits(PQCAException):
    """The partition must contain each qubit exactly once."""

    def __init__(self, cells):
        """Create exception."""
        super().__init__(
            f"Each qubit must appear exactly once in {cells}.")


class EmptyCellException(PQCAException):
    """Each cell must not be empty."""

    def __init__(self):
        """Create exception."""
        super().__init__("Cells cannot be empty.")


class NoCellsException(PQCAException):
    """There must be at least one cell."""

    def __init__(self):
        """Create exception."""
        super().__init__("There must be at least one cell.")


class IrregularCoordinateDimensions(PQCAException):
    """Two coordinates must be of the same dimension to be compared."""

    def __init__(self, coordinate_1, coordinate_2):
        """Create exception."""
        super().__init__(
            f"{coordinate_1} and {coordinate_2} " +
            "must be of the same length, and divide component-wise.")


class IrregularCellSize(PQCAException):
    """Each cell must be the same size."""

    def __init__(self, cells):
        """Create exception."""
        super().__init__(f"Not all cells in {cells} are of the same size.")


class CircuitWrongShapeForCell(PQCAException):
    """The circuit must be applied to qubits 0...(n-1)."""

    def __init__(self, qubits, size_cell):
        """Create exception."""
        super().__init__(f"Could not apply circuit with qubits {qubits} \
            to cell of size {size_cell}")


class BackendError(PQCAException):
    """Pass backend errors through to user."""



"""Test the functionality described in update_frame.py."""

# pylint: disable=import-error
import qiskit
import pytest
import pqca


def test_wind_circuit_around_loop():
    """One gate on each of five cells should be five gates."""
    tes = pqca.tessellation.one_dimensional(10, 2)
    cx_circuit = qiskit.QuantumCircuit(2)
    cx_circuit.cx(0, 1)
    instructions = pqca.update_frame.wind_circuit_around_loop(cx_circuit, tes)
    assert len(instructions) == 5


def test_wind_circuit_around_loop_exceptions():
    """Bad arguments fail informatively."""
    tes = pqca.tessellation.one_dimensional(10, 2)
    cx_circuit = qiskit.QuantumCircuit(3)
    with pytest.raises(pqca.exceptions.CircuitWrongShapeForCell):
        pqca.update_frame.wind_circuit_around_loop(cx_circuit, tes)


def test_update_frame_arguments():
    """Test problematic combinations of arguments."""
    tes = pqca.tessellation.one_dimensional(10, 2)
    with pytest.raises(pqca.exceptions.TooLittleDataForUpdateFrame):
        pqca.update_frame.UpdateFrame(tes)
    with pytest.raises(pqca.exceptions.TooMuchDataForUpdateFrame):
        pqca.update_frame.UpdateFrame(
            tes, qasm_circuit_file="", qasm_data_as_string="")


def test_update_frame_repr():
    """Test the string representation."""
    tes = pqca.tessellation.one_dimensional(10, 2)
    cx_circuit = qiskit.QuantumCircuit(2)
    cx_circuit.cx(0, 1)
    frame = pqca.UpdateFrame(tes, qiskit_circuit=cx_circuit)
    assert len(str(frame)) > 0



"""Partitioned Quantum Cellular Automaton."""

from typing import List, Callable
from qiskit import QuantumCircuit
from .update_frame import UpdateFrame


class Automaton:
    """Partitioned Quantum Cellular Automaton.

    initial_state: a list of 1s and 0s, first bit indicates sate of qubit 0 etc.
    frames: a list of UpdateFrames
    backend: a function that runs a (preparation and update) circuit
    """

    def __init__(self, initial_state: list,
                 frames: List[UpdateFrame], backend: Callable[[QuantumCircuit], List[int]]):
        """PQCA with a given initial state, update frames, and backend.

        The initialisation and update circuit is constructed from the current state
        and the update frames.
        """
        self.frames = frames
        self.backend = backend
        frame_instructions = map(
            lambda f: f.full_circuit_instructions, self.frames)
        self.update_instruction = [instruction for
                                   instructions in frame_instructions for
                                   instruction in instructions]

        size = len(initial_state)

        self.state = initial_state

        self.update_circuit = QuantumCircuit(size)
        for instruction, qargs, cargs in self.update_instruction:
            self.update_circuit.append(instruction, qargs, cargs)

    @property
    def preparation_circuit(self):
        """Circuit for preparing a register of qubits into the current state."""
        return pattern_preparation_circuit(self.state)

    @property
    def combined_circuit(self):
        """Combine preparation and update circuit."""
        return self.preparation_circuit + self.update_circuit

    def tick(self):
        """Update the state without returning anything."""
        assert self.backend, "Backend not yet assigned"
        next_pattern = self.backend(self.combined_circuit)
        self.state = next_pattern

    def iterate(self, number_of_iterations=1):
        """Iterate and return new each state reached."""
        pattern_sequence = []
        for _ in range(0, number_of_iterations):
            self.tick()
            pattern_sequence.append(self.state.copy())
        return pattern_sequence

    def __str__(self):
        """Represent as string."""
        frame_string = f"[{','.join([str(frame) for frame in self.frames])}]"
        return f"PQCA(state={self.state}, frames={frame_string})"


def pattern_preparation_circuit(pattern: list):
    """Create a circuit that encoding the starting state."""
    circuit = QuantumCircuit(len(pattern))
    for (qubit, boolean_value) in list(enumerate(pattern)):
        if boolean_value:
            circuit.x(qubit)
    return circuit



"""Simple n-dimensional, extendable vector of ints."""

from __future__ import annotations
from typing import Callable, List, Tuple


class Vector:
    """Wrap a list of entries with helper functions."""

    entries: List[int]

    def __init__(self, entries: List[int] = None):
        """Convert a tuple of ints into coordinates."""
        self.entries = entries if entries is not None else []

    def extend(self, next_entry: int):
        """Add the given entry to the end of the list of entries."""
        return Vector(list(self.entries) + [next_entry])

    def action(self, other: Vector, action: Callable[[int, int], int]):
        """Apply the action to two vectors."""
        return Vector([action(a, b) for (a, b) in zip(self.entries, other.entries)])

    def __add__(self, other: Vector):
        """Add two vectors."""
        return self.action(other, lambda x, y: x+y)

    def __sub__(self, other: Vector):
        """Subtract two vectors."""
        return self.action(other, lambda x, y: x-y)

    def __eq__(self, other: Vector):
        """Compare component-wise."""
        return self.entries == other.entries

    def __getitem__(self, item):
        """Access to entries."""
        return self.entries[item]

    def __str__(self):
        """Human-readable string representation."""
        return f"Vector{self.entries}"

    def __repr__(self):
        """Unambiguous string representation."""
        return str(self)



"""Test tessellation of circuits."""

# pylint: disable=import-error
import itertools
import re
import pytest
import pqca
from pqca.vector import Vector


def test_create_one_dimensional():
    """Tessellate a (finite) line."""
    tes_one = pqca.tessellation.one_dimensional(10, 2)
    assert str(
        tes_one) == "Tessellation(10 qubits as 5 cells, first cell: [0, 1])"


def test_when_n_is_one():
    """Check that one_dimensional and n_dimensional overlap when n=1."""
    tes_one = pqca.tessellation.one_dimensional(10, 2)
    tes_n = pqca.tessellation.n_dimensional([10], [2])
    assert tes_one.cells == tes_n.cells


def test_vector_to_name():
    """Convert a vector to a point in lexicographic order."""
    vtn = pqca.tessellation.vector_to_name
    assert vtn(Vector([5]), [6]) == 5
    assert vtn(Vector([5, 0]), [6, 6]) == 5*6
    assert vtn(Vector([0, 0]), [6, 6]) == 0
    assert vtn(Vector([0, 5]), [6, 6]) == 5
    assert vtn(Vector([1, 5]), [6, 6]) == 6+5
    assert vtn(Vector([1, 2]), [6, 3]) == 1*3+2
    assert vtn(Vector([0, 2, 0]), [6, 5, 4]) == 2*4
    assert vtn(Vector([1, 2, 3]), [6, 5, 4]) == 1*5*4 + 2*4+3


def test_cell_of_given_size():
    """Test creation of the first cell."""
    cell = pqca.tessellation.cell_of_given_size([2, 3, 2])
    print(cell)
    assert len(cell) == 2*3*2
    assert Vector([0, 0, 0]) in cell
    assert Vector([1, 2, 1]) in cell


def test_create_n_dimensional():
    """Tessellate a (finite) lattice."""
    for dimension in (1, 2, 3, 4):
        total_qubits = pow(6, dimension)
        total_cells = pow(3, dimension)
        tes = pqca.tessellation.n_dimensional([6]*dimension, [2]*dimension)
        assert re.match(f"Tessellation\\({total_qubits} qubits as" +
                        f" {total_cells} cells", str(tes)) is not None


def test_faulty_n_dimensional():
    """Pass bad arguments to n_dimensional."""
    with pytest.raises(pqca.exceptions.IrregularCoordinateDimensions):
        pqca.tessellation.n_dimensional([2], [])
    with pytest.raises(pqca.exceptions.IrregularCoordinateDimensions):
        pqca.tessellation.n_dimensional([2], [3])


def test_faulty_tessellation():
    """Pass bad arguments to the tessellation."""
    with pytest.raises(pqca.exceptions.IrregularCellSize):
        pqca.tessellation.Tessellation([[0], [1, 2]])
    with pytest.raises(pqca.exceptions.PartitionUnevenlyCoversQubits):
        pqca.tessellation.Tessellation([[0], [0]])
    with pytest.raises(pqca.exceptions.EmptyCellException):
        pqca.tessellation.Tessellation([[]])
    with pytest.raises(pqca.exceptions.NoCellsException):
        pqca.tessellation.Tessellation([])


def test_shift():
    """Tessellation.shifted_by."""
    tes_zero = pqca.tessellation.one_dimensional(4, 2)
    tes_one = tes_zero.shifted_by(1)
    assert tes_one.cells == [[1, 2], [3, 0]]
    assert tes_one.shifted_by(-1).cells == tes_zero.cells



"""Test the Vector class."""

# pylint: disable=import-error
import pytest
import pqca
from pqca.vector import Vector


def test_vector_creation():
    """Create and check equality."""
    assert Vector() == Vector()
    assert Vector([1, 2]) == Vector([1, 2])
    assert Vector([1, 2]) != Vector([2, 1])


def test_entry_access():
    """Test the __getitem___."""
    assert Vector([1, 2])[0] == 1


def test_vector_extension():
    """Test the .extend function."""
    assert Vector().extend(1) == Vector([1])
    assert Vector([2]).extend(1) == Vector([2, 1])


def test_vector_string():
    """Test str(Vector)."""
    assert str(Vector([1, 2])) == "Vector[1, 2]"


def test_vector_arithmetic():
    """Test addition and subtraction."""
    assert Vector([1, 1]) + Vector([2, 3]) == Vector([3, 4])
    assert Vector([1, 1]) - Vector([2, 3]) == Vector([-1, -2])
